@@ -1,44 +1,22 @@
 #!/usr/bin/env python3
 #
-# This file is part of Script of Scripts (SoS), a workflow system
-# for the execution of commands and scripts in different languages.
-# Please visit https://github.com/vatlab/SOS for more information.
-#
-# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
-#
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-
+# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
+# Distributed under the terms of the 3-clause BSD License.
 
 import os
-import time
+import subprocess
 import unittest
 from io import StringIO
 
-from sos.sos_script import SoS_Script
+from sos.parser import SoS_Script
+from sos.targets import file_target
 from sos.utils import env
-from sos.sos_executor import Base_Executor
-from sos.target import FileTarget
-
-import subprocess
-import multiprocessing as mp
-try:
-    mp.set_start_method('spawn')
-except:
-    pass
+# if the test is imported under sos/test, test interacive executor
+from sos.workflow_executor import Base_Executor
+from sos import execute_workflow
 
 class TestDAG(unittest.TestCase):
+
     def setUp(self):
         env.reset()
         subprocess.call('sos remove -s', shell=True)
@@ -46,7 +24,8 @@ class TestDAG(unittest.TestCase):
 
     def tearDown(self):
         for f in self.temp_files:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
 
     def touch(self, files):
         '''create temporary files'''
@@ -62,13 +41,23 @@ class TestDAG(unittest.TestCase):
     def assertDAG(self, dag, content):
         if isinstance(dag, str):
             with open(dag) as d:
-                dot = d.read()
+                # only get the first DAG
+                dot = 'strict' + d.read().split('strict')[1]
         else:
             out = StringIO()
-            dag.write_dot(out)
+            dag.save(out)
             dot = out.getvalue()
-        self.assertEqual(sorted([x.strip() for x in dot.split('\n') if x.strip()]),
-            sorted([x.strip() for x in content.split('\n') if x.strip()]))
+        self.assertEqual(
+            sorted([
+                x.strip()
+                for x in dot.split('\n')
+                if x.strip() and not 'digraph' in x
+            ]),
+            sorted([
+                x.strip()
+                for x in content.split('\n')
+                if x.strip() and not 'digraph' in x
+            ]))
 
     def testSimpleDAG(self):
         '''Test DAG with simple dependency'''
@@ -88,8 +77,8 @@ class TestDAG(unittest.TestCase):
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''strict digraph "" {
+        self.assertDAG(
+            dag, '''strict digraph "" {
 A_2;
 A_4;
 A_1;
@@ -114,8 +103,8 @@ input: 'a.txt'
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''strict digraph "" {
+        self.assertDAG(
+            dag, '''strict digraph "" {
 A_2;
 A_4;
 A_1;
@@ -125,7 +114,6 @@ A_3 -> A_4;
 }
 ''')
 
-
         #
         # 1 -> 2 -> 3 -> 4
         #
@@ -149,8 +137,8 @@ output: 'e.txt'
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''strict digraph "" {
+        self.assertDAG(
+            dag, '''strict digraph "" {
 A_2;
 A_4;
 A_1;
@@ -174,7 +162,7 @@ input: 'b.txt'
 output: 'c.txt'
 
 [B_3]
-input: []
+input: None
 output: 'd.txt'
 
 [B_4]
@@ -184,8 +172,8 @@ output: 'e.txt'
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''strict digraph "" {
+        self.assertDAG(
+            dag, '''strict digraph "" {
 B_2;
 B_4;
 B_1;
@@ -219,8 +207,8 @@ output: 'e.txt'
 
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''strict digraph "" {
+        self.assertDAG(
+            dag, '''strict digraph "" {
 B_1;
 B_4;
 B_2;
@@ -254,8 +242,8 @@ output: 'e.txt'
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 C_1;
 C_4;
@@ -298,9 +286,9 @@ output: 'e.txt'
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
         dag.show_nodes()
-        #dag.write_dot('a.dot')
-        self.assertDAG(dag,
-'''
+        # dag.save('a.dot')
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 C_1;
 C_4;
@@ -333,8 +321,8 @@ output: 'e.txt'
         ''')
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 C_1;
 C_4;
@@ -351,10 +339,10 @@ C_3 -> C_4;
     def testAuxiliarySteps(self):
         script = SoS_Script('''
 [K: provides='{name}.txt']
-output: "${name}.txt"
+output: f"{name}.txt"
 
-run:
-    touch '${name}.txt'
+run: expand=True
+    touch '{name}.txt'
 
 [C_2]
 input: 'b.txt'
@@ -377,34 +365,22 @@ input: 'a.txt'
         dag = Base_Executor(wf).initialize_dag()
         #
         # Ticket 363:
-        # 
+        #
         # we have two possibilities here, one is to ignore a.txt,
         # and one is to regenerate a.txt because it is not generated
         # by sos (without signature)        #
         #
-        #dag.write_dot('a.dot')
-        #dag.show_nodes()
-#        self.assertDAG(dag,
-#'''
-#strict digraph "" {
-#"K ['b.txt']";
-#C_3;
-#C_2;
-#"K ['b.txt']" -> C_2;
-#}
-#''')
-        self.assertDAG(dag, '''
-        strict digraph "" {
-"K ['a.txt']";
-"K ['b.txt']";
-C_2;
+        dag.show_nodes()
+        self.assertDAG(
+            dag, '''
+strict digraph "" {
+"K (b.txt)";
 C_3;
-"K ['a.txt']" -> C_3;
-"K ['b.txt']" -> C_2;
+C_2;
+"K (b.txt)" -> C_2;
 }
 ''')
 
-
     def testCycle(self):
         '''Test cycle detection of DAG'''
         #
@@ -432,8 +408,12 @@ output: 'A.txt'
     def testLongChain(self):
         '''Test long make file style dependencies.'''
         #
-        for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            FileTarget(f).remove('both')
+        for f in [
+                'A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt',
+                'C1.txt', 'C3.txt', 'C4.txt'
+        ]:
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         #  A1 <- B1 <- B2 <- B3
         #   |
@@ -493,41 +473,48 @@ run:
         wf = script.workflow()
         #env.verbosity = 4
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
-"C4 ['C4.txt']";
-"B1 ['B1.txt']";
-"C1 ['C1.txt']";
-"C2 ['C2.txt']";
-"C3 ['C3.txt']";
+"C4 (C4.txt)";
+"B1 (B1.txt)";
+"C1 (C1.txt)";
+"C2 (C2.txt)";
+"C3 (C3.txt)";
 A_1;
-"B2 ['B2.txt']";
-"B3 ['B3.txt']";
+"B2 (B2.txt)";
+"B3 (B3.txt)";
 A_2;
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
-"C4 ['C4.txt']" -> "C3 ['C3.txt']";
-"B1 ['B1.txt']" -> A_1;
-"C1 ['C1.txt']" -> "B2 ['B2.txt']";
-"C2 ['C2.txt']" -> "C1 ['C1.txt']";
-"C3 ['C3.txt']" -> "C1 ['C1.txt']";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
+"C4 (C4.txt)" -> "C3 (C3.txt)";
+"B1 (B1.txt)" -> A_1;
+"C1 (C1.txt)" -> "B2 (B2.txt)";
+"C2 (C2.txt)" -> "C1 (C1.txt)";
+"C3 (C3.txt)" -> "C1 (C1.txt)";
 A_1 -> A_2;
-"B2 ['B2.txt']" -> "B1 ['B1.txt']";
-"B2 ['B2.txt']" -> A_2;
-"B3 ['B3.txt']" -> "B2 ['B2.txt']";
+"B2 (B2.txt)" -> "B1 (B1.txt)";
+"B2 (B2.txt)" -> A_2;
+"B3 (B3.txt)" -> "B2 (B2.txt)";
 }
 ''')
         Base_Executor(wf).run()
-        for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            t = FileTarget(f)
-            self.assertTrue(t.exists(), f + ' should exist')
-            t.remove('both')
+        for f in [
+                'A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt',
+                'C1.txt', 'C3.txt', 'C4.txt'
+        ]:
+            t = file_target(f)
+            self.assertTrue(t.target_exists(), f + ' should exist')
+            t.unlink()
 
     def testTarget(self):
         '''Test executing only part of a workflow.'''
         #
-        for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            FileTarget(f).remove('both')
+        for f in [
+                'A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt',
+                'C1.txt', 'C3.txt', 'C4.txt'
+        ]:
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         #  A1 <- B1 <- B2 <- B3
         #   |
@@ -537,13 +524,13 @@ A_1 -> A_2;
         #                    C3
         #
         script = SoS_Script('''
-[A_1]
+[A1]
 input: 'B1.txt'
 output: 'A1.txt'
 run:
     touch A1.txt
 
-[A_2]
+[A2]
 depends:  'B2.txt'
 run:
     touch A2.txt
@@ -583,92 +570,98 @@ run:
 
         ''')
         # the workflow should call step K for step C_2, but not C_3
-        wf = script.workflow()
+        wf = script.workflow(use_default=False)
         #
         # test 1, we only need to generate target 'B1.txt'
         dag = Base_Executor(wf).initialize_dag(targets=['B1.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
-"B3 ['B3.txt']";
-"C4 ['C4.txt']";
-"C2 ['C2.txt']";
-"C1 ['C1.txt']";
-"B1 ['B1.txt']";
-"B2 ['B2.txt']";
-"C3 ['C3.txt']";
-"B3 ['B3.txt']" -> "B2 ['B2.txt']";
-"C4 ['C4.txt']" -> "C3 ['C3.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
-"C2 ['C2.txt']" -> "C1 ['C1.txt']";
-"C1 ['C1.txt']" -> "B2 ['B2.txt']";
-"B2 ['B2.txt']" -> "B1 ['B1.txt']";
-"C3 ['C3.txt']" -> "C1 ['C1.txt']";
+"B3 (B3.txt)";
+"C4 (C4.txt)";
+"C2 (C2.txt)";
+"C1 (C1.txt)";
+"B1 (B1.txt)";
+"B2 (B2.txt)";
+"C3 (C3.txt)";
+"B3 (B3.txt)" -> "B2 (B2.txt)";
+"C4 (C4.txt)" -> "C3 (C3.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
+"C2 (C2.txt)" -> "C1 (C1.txt)";
+"C1 (C1.txt)" -> "B2 (B2.txt)";
+"B2 (B2.txt)" -> "B1 (B1.txt)";
+"C3 (C3.txt)" -> "C1 (C1.txt)";
 }
 ''')
         Base_Executor(wf).run(targets=['B1.txt'])
         for f in ['A1.txt', 'A2.txt']:
-            self.assertFalse(FileTarget(f).exists())
-        for f in ['C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            t = FileTarget(f)
-            self.assertTrue(t.exists())
-            t.remove('both')
+            self.assertFalse(file_target(f).target_exists())
+        for f in [
+                'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt',
+                'C4.txt'
+        ]:
+            t = file_target(f)
+            self.assertTrue(t.target_exists())
+            t.unlink()
         #
         # test 2, we would like to generate two files
         dag = Base_Executor(wf).initialize_dag(targets=['B2.txt', 'C2.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
-"C4 ['C4.txt']";
-"B2 ['B2.txt']";
-"C3 ['C3.txt']";
-"B3 ['B3.txt']";
-"C2 ['C2.txt']";
-"C1 ['C1.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
-"C4 ['C4.txt']" -> "C3 ['C3.txt']";
-"C3 ['C3.txt']" -> "C1 ['C1.txt']";
-"B3 ['B3.txt']" -> "B2 ['B2.txt']";
-"C2 ['C2.txt']" -> "C1 ['C1.txt']";
-"C1 ['C1.txt']" -> "B2 ['B2.txt']";
+"C4 (C4.txt)";
+"B2 (B2.txt)";
+"C3 (C3.txt)";
+"B3 (B3.txt)";
+"C2 (C2.txt)";
+"C1 (C1.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
+"C4 (C4.txt)" -> "C3 (C3.txt)";
+"C3 (C3.txt)" -> "C1 (C1.txt)";
+"B3 (B3.txt)" -> "B2 (B2.txt)";
+"C2 (C2.txt)" -> "C1 (C1.txt)";
+"C1 (C1.txt)" -> "B2 (B2.txt)";
 }
 ''')
         Base_Executor(wf).run(targets=['B2.txt', 'C2.txt'])
         for f in ['A1.txt', 'B1.txt', 'A2.txt']:
-            self.assertFalse(FileTarget(f).exists())
+            self.assertFalse(file_target(f).target_exists())
         for f in ['C2.txt', 'B2.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            t = FileTarget(f)
-            self.assertTrue(t.exists())
-            t.remove('both')
+            t = file_target(f)
+            self.assertTrue(t.target_exists())
+            t.unlink()
         #
         # test 3, generate two separate trees
         #
         dag = Base_Executor(wf).initialize_dag(targets=['B3.txt', 'C2.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
-"B3 ['B3.txt']";
-"C2 ['C2.txt']";
-"C4 ['C4.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
+"B3 (B3.txt)";
+"C2 (C2.txt)";
+"C4 (C4.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
 }
 ''')
         Base_Executor(wf).run(targets=['B3.txt', 'C2.txt'])
         for f in ['A1.txt', 'B1.txt', 'A2.txt', 'B2.txt', 'C1.txt', 'C3.txt']:
-            self.assertFalse(FileTarget(f).exists())
+            self.assertFalse(file_target(f).target_exists())
         for f in ['C2.txt', 'B3.txt', 'C4.txt']:
-            t = FileTarget(f)
-            self.assertTrue(t.exists())
-            t.remove('both')
+            t = file_target(f)
+            self.assertTrue(t.target_exists())
+            t.unlink()
 
     def testPatternReuse(self):
         '''Test repeated use of steps that use pattern and produce different files.'''
         #
-        for f in ['A1.txt', 'A2.txt', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p']:
-            FileTarget(f).remove('both')
+        for f in [
+                'A1.txt', 'A2.txt', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
+        ]:
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         #  A1 <- P <- B1
         #  A1 <- P <- B2
@@ -682,6 +675,7 @@ run:
     touch A1.txt
 
 [A_2]
+output: 'A2.txt'
 run:
     touch A2.txt
 
@@ -695,55 +689,56 @@ run:
 
 [P: provides='{filename}.p']
 input: filename
-run:
-    touch ${output}
+run: expand=True
+    touch {_output}
 ''')
         # the workflow should call step K for step C_2, but not C_3
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
-"P ['B2.txt.p']";
-"B1 ['B1.txt']";
-"B2 ['B2.txt']";
+"P (B2.txt.p)";
+"B1 (B1.txt)";
+"B2 (B2.txt)";
 A_2;
 A_1;
-"P ['B1.txt.p']";
-"P ['B2.txt.p']" -> A_1;
-"B1 ['B1.txt']" -> "P ['B1.txt.p']";
-"B2 ['B2.txt']" -> "P ['B2.txt.p']";
+"P (B1.txt.p)";
+"P (B2.txt.p)" -> A_1;
+"B1 (B1.txt)" -> "P (B1.txt.p)";
+"B2 (B2.txt)" -> "P (B2.txt.p)";
 A_1 -> A_2;
-"P ['B1.txt.p']" -> A_1;
+"P (B1.txt.p)" -> A_1;
 }
 ''')
         Base_Executor(wf).run()
-        for f in ['A1.txt', 'A2.txt', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p']:
-            t = FileTarget(f)
-            self.assertTrue(t.exists())
-            t.remove('both')
-
+        for f in [
+                'A1.txt', 'A2.txt', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
+        ]:
+            t = file_target(f)
+            self.assertTrue(t.target_exists(), '{} should exist'.format(f))
+            t.unlink()
 
     def testParallelExecution(self):
-        '''Test basic parallel execution'''
-        '''
+        '''Test basic parallel execution
         A1 <- None
         A2 <- B2
         '''
         for f in ['A1.txt', 'B2.txt', 'A2.txt']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
         script = SoS_Script('''
 [A_1]
 output: 'A1.txt'
 run:
-    sleep 3
+    sleep 0
     touch A1.txt
 
 [A_2]
 input:  'B2.txt'
 output: 'A2.txt'
 run:
-    sleep 3
+    sleep 0
     touch A2.txt
 
 [B: provides='B2.txt']
@@ -756,31 +751,30 @@ run:
         # the workflow should call step K for step C_2, but not C_3
         wf = script.workflow()
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 A_1;
 A_2;
-"B ['B2.txt']";
-"B ['B2.txt']" -> A_2;
+"B (B2.txt)";
+"B (B2.txt)" -> A_2;
 }
 ''')
         env.max_jobs = 4
-        st = time.time()
         #env.verbosity = 4
         Base_Executor(wf).run()
         # the process is slower after switching to spawn mode
-        self.assertLess(time.time() - st, 7)
         for f in ['A1.txt', 'B2.txt', 'A2.txt']:
-            FileTarget(f).remove('both')
-
+            if file_target(f).exists():
+                file_target(f).unlink()
 
     def testSharedDependency(self):
         #
         # shared variable should introduce additional dependency
         #
         for f in ['A1.txt']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         # A1 introduces a shared variable ss, A3 depends on ss but not A2
         #
@@ -792,20 +786,20 @@ ss = 'A1'
 input: None
 
 run:
-    sleep 3
+    sleep 0
 
 [A_3]
 input: None
 import time
-time.sleep(3)
-with open("${ss}.txt", 'w') as tmp:
+time.sleep(0)
+with open(f"{ss}.txt", 'w') as tmp:
     tmp.write('test')
 
 ''')
         wf = script.workflow('A')
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 A_3;
 A_1;
@@ -814,18 +808,16 @@ A_1 -> A_3;
 }
 ''')
         env.max_jobs = 3
-        st = time.time()
         Base_Executor(wf).run()
-        self.assertLess(time.time() - st, 7)
         for f in ['A1.txt']:
-            self.assertTrue(FileTarget(f).exists())
-            FileTarget(f).remove('both')
-
+            self.assertTrue(file_target(f).target_exists())
+            file_target(f).unlink()
 
     def testLiteralConnection(self):
         '''Testing the connection of steps with by variables.'''
         for f in ['A1.txt']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         # A1 introduces a shared variable ss, A3 depends on ss but not A2
         #
@@ -840,26 +832,29 @@ p = 'A1.txt'
 input: None
 
 run:
-    sleep 3
+    sleep 0
 
 [A_3]
 input: p
+depends: sos_variable('p')
 
 run:
-    sleep 3
+    sleep 0
 
 [A_4]
 input: p
+depends: sos_variable('p')
 run:
-    sleep 3
+    sleep 0
 
 [A_5]
 input: dynamic(p)
+depends: sos_variable('p')
 ''')
         wf = script.workflow('A')
         dag = Base_Executor(wf).initialize_dag()
-        self.assertDAG(dag,
-'''
+        self.assertDAG(
+            dag, '''
 strict digraph "" {
 A_1;
 A_4;
@@ -869,17 +864,13 @@ A_5;
 A_1 -> A_4;
 A_1 -> A_3;
 A_1 -> A_5;
-A_4 -> A_5;
 }
 ''')
         env.max_jobs = 3
-        st = time.time()
         Base_Executor(wf).run()
-        self.assertLess(time.time() - st, 8)
         for f in ['A1.txt']:
-            self.assertTrue(FileTarget(f).exists())
-            FileTarget(f).remove('both')
-
+            self.assertTrue(file_target(f).target_exists())
+            file_target(f).unlink()
 
     def testVariableTarget(self):
         '''Test dependency caused by variable usage.'''
@@ -900,10 +891,10 @@ p = c + b
         Base_Executor(wf).run()
         self.assertTrue(env.sos_dict['p'], 3)
 
-
     def testReverseSharedVariable(self):
         '''Test shared variables defined in auxiliary steps'''
-        FileTarget('a.txt').remove('both')
+        if file_target('a.txt').exists():
+            file_target('a.txt').unlink()
         script = SoS_Script(r'''
 [A: shared='b', provides='a.txt']
 b = 1
@@ -926,30 +917,34 @@ print(b)
         script = SoS_Script(r'''
 # this step provides variable `var`
 [index: provides='{filename}.bam.bai']
-input: "${filename}.bam"
-run:
-   echo "Generating ${output}"
-   touch ${output}
+input: f"{filename}.bam"
+run: expand=True
+   echo "Generating {_output}"
+   touch {_output}
 
 [call: provides='{filename}.vcf']
-input:   "${filename}.bam"
-depends: "${input}.bai"
-run:
-   echo "Calling variants from ${input} with ${depends} to ${output}"
-   touch ${output}
+input:   f"{filename}.bam"
+depends: f"{_input}.bai"
+run: expand=True
+   echo "Calling variants from {_input} with {_depends} to {_output}"
+   touch {_output}
 ''')
-        FileTarget('a.bam.bai').remove('both')
-        FileTarget('a.vcf').remove('both')
+        if file_target('a.bam.bai').exists():
+            file_target('a.bam.bai').unlink()
+        if file_target('a.vcf').exists():
+            file_target('a.vcf').unlink()
         self.touch('a.bam')
         Base_Executor(script.workflow()).run(targets=['a.vcf'])
-        for file in ('a.vcf', 'a.bam', 'a.bam.bai'):
-            FileTarget(file).remove('both')
+        for f in ('a.vcf', 'a.bam', 'a.bam.bai'):
+            if file_target(f).exists():
+                file_target(f).unlink()
 
     def testOutputOfDAG(self):
         '''Test output of dag'''
         #
-        for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
-            FileTarget(f).remove('both')
+        #for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
+        #    if file_target(f).exists():
+        #        file_target(f).unlink()
         #
         #  A1 <- B1 <- B2 <- B3
         #   |
@@ -1005,65 +1000,78 @@ run:
 
         ''')
         # the workflow should call step K for step C_2, but not C_3
-        wf = script.workflow()
+        wf = script.workflow(use_default=False)
         #
         # test 1, we only need to generate target 'B1.txt'
-        Base_Executor(wf, config={'output_dag':'test'}).initialize_dag(targets=['B1.txt'])
+        Base_Executor(
+            wf, config={
+                'output_dag': 'test_outofdag1.dot',
+                'trace_existing': True
+            }).initialize_dag(targets=['B1.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG('test_2.dot',
-'''
+        self.assertDAG(
+            'test_outofdag1.dot', '''
 strict digraph "" {
-"B3 ['B3.txt']";
-"C4 ['C4.txt']";
-"C2 ['C2.txt']";
-"C1 ['C1.txt']";
-"B1 ['B1.txt']";
-"B2 ['B2.txt']";
-"C3 ['C3.txt']";
-"B3 ['B3.txt']" -> "B2 ['B2.txt']";
-"C4 ['C4.txt']" -> "C3 ['C3.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
-"C2 ['C2.txt']" -> "C1 ['C1.txt']";
-"C1 ['C1.txt']" -> "B2 ['B2.txt']";
-"B2 ['B2.txt']" -> "B1 ['B1.txt']";
-"C3 ['C3.txt']" -> "C1 ['C1.txt']";
+"B3 (B3.txt)";
+"C4 (C4.txt)";
+"C2 (C2.txt)";
+"C1 (C1.txt)";
+"B1 (B1.txt)";
+"B2 (B2.txt)";
+"C3 (C3.txt)";
+"B3 (B3.txt)" -> "B2 (B2.txt)";
+"C4 (C4.txt)" -> "C3 (C3.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
+"C2 (C2.txt)" -> "C1 (C1.txt)";
+"C1 (C1.txt)" -> "B2 (B2.txt)";
+"B2 (B2.txt)" -> "B1 (B1.txt)";
+"C3 (C3.txt)" -> "C1 (C1.txt)";
 }
 ''')
         # test 2, we would like to generate two files
-        Base_Executor(wf, config={'output_dag':'test'}).initialize_dag(targets=['B2.txt', 'C2.txt'])
+        Base_Executor(
+            wf, config={
+                'output_dag': 'test_outofdag2.dot',
+                'trace_existing': True
+            }).initialize_dag(targets=['B2.txt', 'C2.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG('test_2.dot',
-'''
+        self.assertDAG(
+            'test_outofdag2.dot', '''
 strict digraph "" {
-"C4 ['C4.txt']";
-"B2 ['B2.txt']";
-"C3 ['C3.txt']";
-"B3 ['B3.txt']";
-"C2 ['C2.txt']";
-"C1 ['C1.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
-"C4 ['C4.txt']" -> "C3 ['C3.txt']";
-"C3 ['C3.txt']" -> "C1 ['C1.txt']";
-"B3 ['B3.txt']" -> "B2 ['B2.txt']";
-"C2 ['C2.txt']" -> "C1 ['C1.txt']";
-"C1 ['C1.txt']" -> "B2 ['B2.txt']";
+"C4 (C4.txt)";
+"B2 (B2.txt)";
+"C3 (C3.txt)";
+"B3 (B3.txt)";
+"C2 (C2.txt)";
+"C1 (C1.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
+"C4 (C4.txt)" -> "C3 (C3.txt)";
+"C3 (C3.txt)" -> "C1 (C1.txt)";
+"B3 (B3.txt)" -> "B2 (B2.txt)";
+"C2 (C2.txt)" -> "C1 (C1.txt)";
+"C1 (C1.txt)" -> "B2 (B2.txt)";
 }
 ''')
         # test 3, generate two separate trees
         #
-        Base_Executor(wf, config={'output_dag':'test'}).initialize_dag(targets=['B3.txt', 'C2.txt'])
+        Base_Executor(
+            wf, config={
+                'output_dag': 'test_outofdag3.dot',
+                'trace_existing': True
+            }).initialize_dag(targets=['B3.txt', 'C2.txt'])
         # note that A2 is no longer mentioned
-        self.assertDAG('test_2.dot',
-'''
+        self.assertDAG(
+            'test_outofdag3.dot', '''
 strict digraph "" {
-"B3 ['B3.txt']";
-"C2 ['C2.txt']";
-"C4 ['C4.txt']";
-"C4 ['C4.txt']" -> "C2 ['C2.txt']";
+"B3 (B3.txt)";
+"C2 (C2.txt)";
+"C4 (C4.txt)";
+"C4 (C4.txt)" -> "C2 (C2.txt)";
 }
 ''')
         for f in ['C2.txt', 'B3.txt', 'C4.txt', 'test.dot', 'test_2.dot']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
 
     def testStepWithMultipleOutput(self):
         '''Test addition of steps with multiple outputs. It should be added only once'''
@@ -1071,23 +1079,220 @@ strict digraph "" {
 [test_1: provides=['{}.txt'.format(i) for i in range(10)]]
 output: ['{}.txt'.format(i) for i in range(10)]
 run:
-  touch ${output}
+  touch {output}
 
 [test_2: provides=['{}.txt'.format(i) for i in range(10, 20)]]
 depends: ['{}.txt'.format(i) for i in range(10)]
 output: ['{}.txt'.format(i) for i in range(10, 20)]
 run:
-  touch ${output}
+  touch {output}
 
 [default]
 depends: ['{}.txt'.format(i) for i in range(10, 20)]
 ''')
         wf = script.workflow()
-        Base_Executor(wf, config={'output_dag':'test'}).initialize_dag()
+        Base_Executor(wf, config={'output_dag': 'test.dot'}).initialize_dag()
         with open('test.dot') as dot:
             lc = len(dot.readlines())
         self.assertTrue(lc, 6)
+        file_target('test.dot').unlink()
 
+    def testAuxiliarySosStep(self):
+        '''Testing the use of sos_step with auxiliary step. #736'''
+        script = SoS_Script('''
+[default]
+depends: '1.txt'
+
+[A_1]
+print("Hi")
+
+
+[C_1: provides = "1.txt"]
+depends: sos_step("A_1")
+run:
+touch 1.txt
+''')
+        wf = script.workflow()
+        Base_Executor(wf).run()
+
+    def testForwardStyleDepend(self):
+        '''Test the execution of forward-style workflow with undtermined dependency'''
+        if file_target('a.txt.bak').exists():
+            file_target('a.txt.bak').unlink()
+        self.touch('a.txt')
+        script = SoS_Script('''
+[10]
+input: 'a.txt'
+output: f"{_input}.bak"
+run: expand=True
+    cp {_input} {_output}
+
+[20]
+depends: "a.txt.bak"
+run: expand=True
+    ls {_depends}
+''')
+        wf = script.workflow()
+        Base_Executor(wf).run()
+        self.assertTrue(file_target('a.txt.bak').target_exists())
+
+    def testSoSStepMiniworkflow(self):
+        '''Test the addition of mini forward workflows introduced by sos_step'''
+        script = SoS_Script('''
+[a_1]
+print(step_name)
+
+[a_2]
+print(step_name)
+[a_20]
+print(step_name)
+
+[b_1]
+print(step_name)
+
+[b_2]
+print(step_name)
+
+[b_20]
+depends: sos_step('c')
+print(step_name)
+
+[c_1]
+print(step_name)
+
+[c_2]
+print(step_name)
+
+[c_20]
+print(step_name)
+
+
+
+[default]
+depends: sos_step('a'), sos_step('b')
+''')
+        wf = script.workflow()
+        Base_Executor(wf, config={'output_dag': 'test.dot'}).run()
+        # note that A2 is no longer mentioned
+        self.assertDAG(
+            'test.dot', '''
+strict digraph "" {
+default;
+a_1;
+a_2;
+a_20;
+b_1;
+b_2;
+b_20;
+c_1;
+c_2;
+c_20;
+a_1 -> a_2;
+a_2 -> a_20;
+a_20 -> default;
+b_1 -> b_2;
+b_2 -> b_20;
+b_20 -> default;
+c_1 -> c_2;
+c_2 -> c_20;
+c_20 -> b_20;
+}
+''')
+        file_target('test.dot').unlink()
+
+    def testMultiNamedOutput(self):
+        '''Test DAG built from multiple named_output #1166'''
+        script = SoS_Script('''
+[A]
+output: A='a.txt', B='b.txt'
+_output.touch()
+
+[default]
+input: named_output('A'), named_output('B')
+''')
+        wf = script.workflow()
+        Base_Executor(wf, config={'output_dag': 'test_named_output.dot'}).run()
+        # note that A2 is no longer mentioned
+        self.assertDAG(
+            'test_named_output.dot', '''
+strict digraph "" {
+default;
+"A (B)";
+"A (B)" -> default;
+}
+''')
+
+    def testCompoundWorkflow(self):
+        '''Test the DAG of compound workflow'''
+        script = SoS_Script('''
+[A_1]
+[A_2]
+[B]
+ ''')
+        wf = script.workflow('A+B')
+        dag = Base_Executor(wf).initialize_dag()
+        self.assertDAG(
+            dag, '''strict digraph "" {
+A_1;
+A_2;
+B;
+A_1 -> A_2;
+A_2 -> B;
+}''')
+        # with empty depends
+        script = SoS_Script('''
+[A_1]
+[A_2]
+[B]
+depends:
+ ''')
+        wf = script.workflow('A+B')
+        dag = Base_Executor(wf).initialize_dag()
+        self.assertDAG(
+            dag, '''strict digraph "" {
+A_1;
+A_2;
+B;
+A_1 -> A_2;
+A_2 -> B;
+}''')
+        if os.path.isfile('a.txt'):
+            os.remove('a.txt')
+        script = SoS_Script('''
+[A_1]
+[A_2]
+[C]
+output: 'a.txt'
+_output.touch()
+
+[B]
+depends: 'a.txt'
+ ''')
+        # with more depends
+        wf = script.workflow('A+B')
+        dag = Base_Executor(wf).initialize_dag()
+        self.assertDAG(
+            dag, '''strict digraph "" {
+A_1;
+A_2;
+B;
+"C (a.txt)";
+A_1 -> A_2;
+A_2 -> B;
+"C (a.txt)" -> B;
+}''')
+
+
+    def testProvidesSoSVariable(self):
+        '''Test provides non-filename targets #1341'''
+        execute_workflow('''
+[count: provides=sos_variable('numNotebooks')]
+numNotebooks = 1
+
+[default]
+depends: sos_variable('numNotebooks')
+print(f"There are {numNotebooks} notebooks in this directory")
+        ''')
 
 if __name__ == '__main__':
     unittest.main()
