@@ -1,36 +1,20 @@
 #!/usr/bin/env python3
 #
-# This file is part of Script of Scripts (SoS), a workflow system
-# for the execution of commands and scripts in different languages.
-# Please visit https://github.com/vatlab/SOS for more information.
-#
-# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
-#
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-
+# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
+# Distributed under the terms of the 3-clause BSD License.
 import os
-import unittest
 import shutil
 import subprocess
+import unittest
 
+from sos.parser import SoS_Script
+from sos.targets import file_target
 from sos.utils import env
-from sos.sos_script import SoS_Script
-from sos.sos_executor import Base_Executor, ExecuteError
-from sos.target import FileTarget
+from sos.workflow_executor import Base_Executor
+from sos import execute_workflow
 
 class TestNested(unittest.TestCase):
+
     def setUp(self):
         env.reset()
         subprocess.call('sos remove -s', shell=True)
@@ -38,7 +22,8 @@ class TestNested(unittest.TestCase):
 
     def tearDown(self):
         for f in self.temp_files:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
 
     def touch(self, files):
         '''create temporary files'''
@@ -55,7 +40,7 @@ class TestNested(unittest.TestCase):
         # progress bar with nested workflow
         script = SoS_Script('''
 import time
-time.sleep(0.5)
+time.sleep(0)
 [sub_1]
 [sub_2]
 [sub_3]
@@ -82,46 +67,59 @@ if 'inputs' not in locals():
 
 [a_1: shared=['executed', 'inputs']]
 executed.append(step_name)
-inputs.append(input)
+inputs.append(_input)
+
 [a_2: shared=['executed', 'inputs']]
 executed.append(step_name)
-inputs.append(input)
+inputs.append(_input)
+
 [a_3: shared=['executed', 'inputs']]
 executed.append(step_name)
-inputs.append(input)
+inputs.append(_input)
+
 [a_4: shared=['executed', 'inputs']]
 executed.append(step_name)
 output: 'a.done'
-inputs.append(input)
-sh:
-    touch ${output}
+inputs.append(_input)
+run: expand=True
+  touch {_output}
+
 [b_1: shared=['executed', 'inputs']]
 executed.append(step_name)
 input: 'b.begin'
-inputs.append(input)
+inputs.append(_input)
+
 [b_2: shared=['executed', 'inputs']]
 executed.append(step_name)
-inputs.append(input)
+inputs.append(_input)
+
 [b_3: shared=['executed', 'inputs']]
 executed.append(step_name)
-inputs.append(input)
+inputs.append(_input)
+
 [b_4: shared=['executed', 'inputs']]
 executed.append(step_name)
 output: 'b.txt'
-inputs.append(input)
+inputs.append(_input)
+
 [c: shared=['executed', 'inputs']]
 executed.append(step_name)
 input: 'a.txt'
 output: 'b.txt'
-inputs.append(input)
+inputs.append(_input)
 sos_run('a+b', shared=['executed', 'inputs'])
 ''')
         env.config['sig_mode'] = 'ignore'
         wf = script.workflow('c')
         Base_Executor(wf).run()
         # order of execution is not guaranteed
-        self.assertEqual(sorted(env.sos_dict['executed']), sorted(['c_0', 'a_1', 'a_2', 'a_3', 'a_4',
-            'b_1', 'b_2', 'b_3', 'b_4']))
+        self.assertEqual(
+            sorted(env.sos_dict['executed']),
+            sorted(
+                ['c', 'a_1', 'a_2', 'a_3', 'a_4', 'b_1', 'b_2', 'b_3', 'b_4']))
+        env.sos_dict.pop('executed', None)
+
+    def testLoopedNestedWorkflow(self):
         # step will be looped
         self.touch(['a.txt', 'b.txt'])
         script = SoS_Script('''
@@ -133,15 +131,17 @@ if 'inputs' not in locals():
 [a_1:shared=['executed', 'inputs']]
 executed.append(step_name)
 output: _input[0] + '.a1'
-inputs.append(input)
-sh:
-    touch ${output}
+inputs.append(_input)
+run: expand=True
+    touch {_output}
+
 [a_2:shared=['executed', 'inputs']]
 executed.append(step_name)
 output: _input[0] + '.a2'
-inputs.append(input)
-sh:
-    touch ${output}
+inputs.append(_input)
+run: expand=True
+    touch {_output}
+
 [c:shared=['executed', 'inputs']]
 executed.append(step_name)
 input: 'a.txt', 'b.txt', group_by='single'
@@ -150,11 +150,17 @@ sos_run('a', shared=['executed', 'inputs'])
 ''')
         wf = script.workflow('c')
         Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['executed'], ['c_0', 'a_1', 'a_2', 'a_1', 'a_2'])
+        self.assertEqual(env.sos_dict['executed'],
+                         ['c', 'a_1', 'a_2', 'a_1', 'a_2'])
         #self.assertEqual(env.sos_dict['inputs'], [['a.txt'], ['a.txt'], ['a.txt.a1'], ['b.txt'], ['b.txt'], ['b.txt.a1']])
         for file in ('a.txt.a1', 'a.txt.a1.a2', 'b.txt.a1', 'b.txt.a1.a2'):
-            FileTarget(file).remove('both')
+            if file_target(file).exists():
+                file_target(file).unlink()
         #
+
+    def testSingleLoopedNestedWorkflow(self):
+        self.touch(['a.txt', 'b.txt'])
+        env.sos_dict.pop('executed', None)
         # allow specifying a single step
         # step will be looped
         script = SoS_Script('''
@@ -167,13 +173,15 @@ executed.append(step_name)
 [c_0:shared='executed']
 executed.append(step_name)
 [c_1:shared='executed']
+depends: sos_variable('executed')
 executed.append(step_name)
 input: 'a.txt', 'b.txt', group_by='single'
 sos_run('a:2', shared='executed')
 ''')
         wf = script.workflow('c')
         Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_2',  'a_2'])
+        self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_2', 'a_2'])
+        env.sos_dict.pop('executed', None)
         # allow specifying a single step
         # step will be looped
         script = SoS_Script('''
@@ -186,6 +194,7 @@ executed.append(step_name)
 [c_0:shared='executed']
 executed.append(step_name)
 [c_1:shared='executed']
+depends: sos_variable('executed')
 executed.append(step_name)
 input: 'a.txt', 'b.txt', group_by='single'
 sos_run('a:2', shared='executed')
@@ -194,7 +203,11 @@ sos_run('a:2', shared='executed')
         Base_Executor(wf).run()
         self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_2', 'a_2'])
         #
+        env.sos_dict.pop('executed', None)
+
+    def testRecursiveNestedWorkflow(self):
         # recursive subworkflow not allowed
+        self.touch(['a.txt', 'b.txt'])
         script = SoS_Script('''
 if 'executed' not in locals():
     executed = []
@@ -210,8 +223,9 @@ input: 'a.txt', 'b.txt', group_by='single'
 sos_run('a_2+c', shared='executed')
 ''')
         wf = script.workflow('c')
-        self.assertRaises(ExecuteError, Base_Executor(wf).run)
+        self.assertRaises(Exception, Base_Executor(wf).run)
         #
+        env.sos_dict.pop('executed', None)
         # nested subworkflow is allowed
         script = SoS_Script('''
 if 'executed' not in locals():
@@ -230,17 +244,23 @@ sos_run('a:1-2', shared='executed')
 [c_0:shared='executed']
 executed.append(step_name)
 [c_1:shared='executed']
+depends: sos_variable('executed')
 executed.append(step_name)
 input: 'a.txt'
 sos_run('a+b', shared='executed')
 ''')
         wf = script.workflow('c')
         Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_1', 'a_2', 'a_3',
-            'b_1', 'b_2', 'a_1', 'a_2'])
+        self.assertEqual(
+            env.sos_dict['executed'],
+            ['c_0', 'c_1', 'a_1', 'a_2', 'a_3', 'b_1', 'b_2', 'a_1', 'a_2'])
         #
         #
+        env.sos_dict.pop('executed', None)
+
+    def testSubworkflowWithOptions(self):
         # nested subworkflow with step option and others
+        self.touch(['a.txt', 'b.txt'])
         script = SoS_Script('''
 if 'executed' not in locals():
     executed = []
@@ -264,16 +284,19 @@ input: 'a.txt', 'b.txt', group_by='single'
 ''')
         wf = script.workflow('b')
         Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['executed'], ['b_0', 'a_3', 'a_1', 'a_3', 'a_1'])
+        self.assertEqual(env.sos_dict['executed'],
+                         ['b', 'a_3', 'a_1', 'a_3', 'a_1'])
+        env.sos_dict.pop('executed', None)
         wf = script.workflow('d')
         Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['executed'], ['d_0', 'a_2', 'a_2'])
+        self.assertEqual(env.sos_dict['executed'], ['d', 'a_2', 'a_2'])
+        env.sos_dict.pop('executed', None)
         wf = script.workflow('e2')
         Base_Executor(wf).run()
         self.assertEqual(env.sos_dict['executed'], ['e2_2'])
         #
         # clean up
-        FileTarget('a.done').remove('both')
+        file_target('a.done').unlink()
 
     def testDynamicNestedWorkflow(self):
         '''Test nested workflow controlled by command line option'''
@@ -301,85 +324,24 @@ sos_run(wf, shared='executed')
 ''')
         wf = script.workflow()
         Base_Executor(wf, args=['--wf', 'b']).run()
-        self.assertEqual(env.sos_dict['executed'], [ 'default_0', 'b_1', 'b_2', 'b_3'])
+        self.assertEqual(env.sos_dict['executed'],
+                         ['default', 'b_1', 'b_2', 'b_3'])
         #
+        env.sos_dict.pop('executed', None)
         Base_Executor(wf, args=['--wf', 'a']).run()
-        self.assertEqual(env.sos_dict['executed'], ['default_0', 'a_1', 'a_2', 'a_3'])
-
-    def testIncludedNestedWorkFlow(self):
-        '''Test the source option of sos_run'''
-        # nested subworkflow with step option and others
-        self.touch(['a.txt', 'b.txt'])
-        #
-        with open('inc.sos', 'w') as sos:
-            sos.write('''
-# test sos script
-
-# global definition
-GLB = 5
-parameter: parB = 10
-
-[A_1: shared='executed']
-executed.append('t.' + step_name)
-output: _input[0] + '.a1'
-sh:
-    touch ${output}
-
-[A_2: shared='executed']
-executed.append('t.' + step_name)
-output: _input[0] + '.a2'
-sh:
-    touch ${output}
-''')
-        script = SoS_Script('''
-%from inc include *
-
-if 'executed' not in locals():
-    executed = []
-
-[b_1: skip=False, shared='executed']
-executed.append(step_name)
-input: 'a.txt', 'b.txt', group_by='single'
-sos_run('A', shared='executed')
-''')
-        wf = script.workflow('b')
-        Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['GLB'], 5)
-        self.assertEqual(env.sos_dict['parB'], 10)
-        self.assertEqual(env.sos_dict['executed'], ['b_1', 't.A_1', 't.A_2', 't.A_1', 't.A_2'])
-        #
-        subprocess.call('sos remove -s', shell=True)
-        for file in ('a.txt.a1', 'a.txt.a1.a2', 'b.txt.a1', 'b.txt.a1.a2'):
-            FileTarget(file).remove('both')
-        script = SoS_Script('''
-%include inc as k
-
-if 'executed' not in locals():
-    executed = []
-
-[b_1: skip=False, shared='executed']
-executed.append('g.' + step_name)
-input: 'a.txt', 'b.txt', group_by='single'
-sos_run('k.A', shared='executed')
-''')
-        wf = script.workflow('b')
-        Base_Executor(wf).run()
-        self.assertEqual(env.sos_dict['k'].GLB, 5)
-        self.assertEqual(env.sos_dict['k'].parB, 10)
-        self.assertEqual(env.sos_dict['executed'], ['g.b_1', 't.k.A_1', 't.k.A_2', 't.k.A_1', 't.k.A_2'])
-        #
-        os.remove('inc.sos')
-
+        self.assertEqual(env.sos_dict['executed'],
+                         ['default', 'a_1', 'a_2', 'a_3'])
 
     def testSoSRun(self):
         '''Test action sos_run with keyword parameters'''
         for f in ['0.txt', '1.txt']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
         script = SoS_Script(r'''
 [A]
 parameter: num=5
-sh:
-    touch ${num}.txt
+run: expand=True
+    touch {num}.txt
 
 [batch]
 for k in range(2):
@@ -388,17 +350,18 @@ for k in range(2):
         wf = script.workflow('batch')
         Base_Executor(wf).run()
         for f in ['0.txt', '1.txt']:
-            self.assertTrue(FileTarget(f).exists())
-            FileTarget(f).remove('both')
+            self.assertTrue(file_target(f).target_exists())
+            file_target(f).unlink()
         #
         # if we do not pass num, parameter would not change
         for f in ['0.txt', '1.txt']:
-            FileTarget(f).remove('both')
+            if file_target(f).exists():
+                file_target(f).unlink()
         script = SoS_Script(r'''
 [A]
 parameter: num=5
-sh:
-    touch ${num}.txt
+run: expand=True
+    touch {num}.txt
 
 [batch]
 for num in range(2):
@@ -407,12 +370,12 @@ for num in range(2):
         wf = script.workflow('batch')
         Base_Executor(wf).run()
         for f in ['0.txt', '1.txt']:
-            self.assertFalse(FileTarget(f).exists())
-        self.assertTrue(FileTarget('5.txt').exists())
-        FileTarget('5.txt').remove('both')
+            self.assertFalse(file_target(f).target_exists())
+        self.assertTrue(file_target('5.txt').target_exists())
+        file_target('5.txt').unlink()
         #
         # test parameter shared to send and return vars
-        # 
+        #
         script = SoS_Script(r'''
 [A: shared='k']
 k += 10
@@ -420,13 +383,13 @@ k += 10
 [batch]
 for k in range(2):
     sos_run('A', shared='k')
-    sh("touch ${k}.txt")
+    run(f"touch {k}.txt")
 ''')
         wf = script.workflow('batch')
         Base_Executor(wf).run()
         for f in ['10.txt', '11.txt']:
-            self.assertTrue(FileTarget(f).exists())
-            FileTarget(f).remove('both')
+            self.assertTrue(file_target(f).target_exists())
+            file_target(f).unlink()
 
     def testDAGofDynamicNestedWorkflow(self):
         #
@@ -434,8 +397,11 @@ for k in range(2):
         # until run time, the DAG should not contain nested workflow
         # until runtime.
         #
-        for f in ['B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p']:
-            FileTarget(f).remove('both')
+        for f in [
+                'B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
+        ]:
+            if file_target(f).exists():
+                file_target(f).unlink()
         #
         #  A1 <- P <- B
         #  A1 <- P <- B
@@ -446,16 +412,16 @@ for k in range(2):
         script = SoS_Script('''
 [A_1]
 parameter: num = 2
-input: "B${num}.txt.p"
+input: f"B{num}.txt.p"
 
 [B: provides='B{num}.txt']
-sh:
-    touch 'B${num}.txt'
+run: expand=True
+    touch 'B{num[0]}.txt'
 
 [P: provides='{filename}.p']
 input: filename
-sh:
-    touch ${output}
+run: expand=True
+    touch {_output}
 
 [ALL]
 
@@ -467,30 +433,44 @@ for i in range(3):
         # the workflow should call step K for step C_2, but not C_3
         wf = script.workflow('ALL')
         Base_Executor(wf).run()
-        for f in ['B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p']:
-            self.assertTrue(FileTarget(f).exists())
-            FileTarget(f).remove('both')
+        for f in [
+                'B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
+        ]:
+            self.assertTrue(file_target(f).target_exists())
+            file_target(f).unlink()
 
+    def testOutcomeOrientedNestedWorkflow(self):
+        '''test nested workflow triggered by targets'''
+        if os.path.isfile('test_15.txt'):
+            os.remove('test_15.txt')
+        execute_workflow('''
+[A: provides='test_{idx}.txt']
+_output.touch()
+
+[default]
+sos_run(targets='test_15.txt')
+        ''')
+        self.assertTrue(os.path.isfile('test_15.txt'))
 
     def testPassingVarsToNestedWorkflow(self):
         '''Test if variables can be passed to nested workflows'''
         script = SoS_Script(r"""
-%set_options sigil='[ ]'
 import time
 import random
 
 [nested]
-print('I am nested [nested] with seed [seed]')
+parameter: nested=True
+parameter: seed=1
+print(f'I am nested {nested} with seed {seed}')
 
 [0]
 reps = range(5)
 input: for_each='reps'
-task: concurrent=True
 import random
 nested = _reps
 seed = random.randint(1, 1000)
-print('Passing [seed] to [nested]')
-sos_run('nested')
+print(f'Passing {seed} to {nested}')
+sos_run('nested', nested=nested, seed=seed)
 
 """)
         wf = script.workflow()
@@ -503,7 +483,7 @@ sos_run('nested')
 def myfunc():
   return 'a'
 
-[1: shared={'test':'output'}]
+[1: shared={'test':'_output'}]
 output: myfunc()
 
 myfunc()
@@ -522,7 +502,7 @@ def myfunc():
     # test if builtin functions (sum and range) can be used here.
     return 'a' + str(sum(range(10)))
 
-[1: shared={'test':'output'}]
+[1: shared={'test':'_output'}]
 output: [myfunc() for i in range(10)][0]
 
 myfunc()
@@ -532,7 +512,6 @@ myfunc()
         Base_Executor(wf).run(mode='dryrun')
         self.assertEqual(env.sos_dict['test'], ['a45'])
 
-
     def testConfigFileOfNestedWorkflow(self):
         '''Test passing of configurationg to nested workflow'''
         script = SoS_Script('''
@@ -548,7 +527,6 @@ sos_run('test:1', key = '1')
         wf = script.workflow()
         Base_Executor(wf, config={'config_file': 'test.conf'}).run()
 
-
     def testErrorFromSubworkflow(self):
         '''Test if error from subworkflow is passed to master (#396)'''
         script = SoS_Script('''
@@ -564,12 +542,13 @@ sos_run('test')
 
     def testFunDef(self):
         '''Test defintion of function that can be used by other steps'''
+        self.touch(['aa.txt', 'ab.txt'])
         # in nested workflow?
         script = SoS_Script(r"""
 def myfunc(a):
     return ['a' + x for x in a]
 
-[mse: shared={'test':'output'}]
+[mse: shared={'test':'_output'}]
 input: myfunc(['a.txt', 'b.txt'])
 
 [1]
@@ -583,28 +562,27 @@ sos_run('mse')
 
     def testSearchPath(self):
         '''Test if any action should exit in five seconds in dryrun mode'''
-        sos_config_file = 'config.yml'
+        sos_config_file = os.path.join(
+            os.path.expanduser('~'), '.sos', 'config.yml')
+        shutil.copy(sos_config_file, 'test.yml')
         #
-        with open(sos_config_file, 'w') as sos_config:
-            sos_config.write('''
-#
-# global sos configuration file
-#
-{{
-    "sos_path": ["{0}/crazy_path", "{0}/crazy_path/more_crazy/"]
-}}
-'''.format(os.getcwd()))
+        subprocess.call(
+            'sos config --set sos_path {0}/crazy_path {0}/crazy_path/more_crazy/'
+            .format(os.getcwd()),
+            shell=True)
         #
         if not os.path.isdir('crazy_path'):
             os.mkdir('crazy_path')
-            os.mkdir('crazy_path/more_crazy')
-        with open('crazy_path/crazy_master.sos', 'w') as crazy:
+            os.mkdir(os.path.join('crazy_path', 'more_crazy'))
+        with open(os.path.join('crazy_path', 'crazy_master.sos'), 'w') as crazy:
             crazy.write('''
 [0]
 sos_run('cc', source='crazy_slave.sos')
 
 ''')
-        with open('crazy_path/more_crazy/crazy_slave.sos', 'w') as crazy:
+        with open(
+                os.path.join('crazy_path', 'more_crazy', 'crazy_slave.sos'),
+                'w') as crazy:
             crazy.write('''
 [cc_0]
 print('hay, I am crazy')
@@ -615,7 +593,7 @@ print('hay, I am crazy')
         #
         shutil.rmtree('crazy_path')
         os.remove(sos_config_file)
-
+        shutil.copy('test.yml', sos_config_file)
 
     def testNestedWorkdir(self):
         '''Test nested runtime option for work directory'''
@@ -624,23 +602,181 @@ print('hay, I am crazy')
         script = SoS_Script('''
 [step]
 task: workdir='tmp'
-bash:
+run:
     touch 'a.txt'
 
 [default]
-task: workdir='tmp'
-sos_run('step')
+sos_run('step', workdir='tmp')
+''')
+        wf = script.workflow()
+        # this should be ok.
+        Base_Executor(wf, config={'default_queue': 'localhost'}).run()
+        os.path.isfile('tmp/tmp/a.txt')
+        shutil.rmtree('tmp')
+
+    def testFailureOfNestedWorkflow(self):
+        '''Test failure of nested workflow #838'''
+        if os.path.isfile('a.txt'):
+            os.remove('a.txt')
+        script = SoS_Script('''
+[something]
+input: 'a.txt'
+
+[default]
+sos_run('something')
+''')
+        wf = script.workflow()
+        # this should be ok.
+        self.assertRaises(Exception, Base_Executor(wf).run)
+
+    def testNestedFromAnotherFile(self):
+        '''Test nested runtime option for work directory'''
+        if os.path.isdir('a.txt'):
+            os.remove('a.txt')
+        with open('another.sos', 'w') as another:
+            another.write('''
+[whatever]
+run:
+    touch 'a.txt'
+
+''')
+        script = SoS_Script('''
+[default]
+sos_run('whatever', source='another.sos')
 ''')
         wf = script.workflow()
         # this should be ok.
         Base_Executor(wf).run()
-        os.path.isfile('tmp/tmp/a.txt')
-        shutil.rmtree('tmp')
+        self.assertTrue(
+            os.path.isfile('a.txt'),
+            'a.txt should have been created by nested workflow from another file'
+        )
+
+    def testConcurrentSubWorkflow(self):
+        '''Test concurrent subworkflow sos_run '''
+        script = SoS_Script('''
+[A]
+parameter: idx=0
+import time
+time.sleep(5)
+
+[default]
+input: for_each=dict(i=range(6))
+sos_run('A', idx=i)
+''')
+        import time
+        st = time.time()
+        wf = script.workflow()
+        # this should be ok.
+        Base_Executor(wf, config={'worker_procs': ['8']}).run()
+        self.assertTrue(time.time() - st < 30)
+
+    def testSoSMultiWorkflow(self):
+        '''Test multiple workflows in sos_run '''
+        script = SoS_Script('''
+[B]
+parameter: idx=2
+import time
+time.sleep(idx)
+
+[A]
+import time
+time.sleep(idx)
+
+[default]
+input: for_each=dict(i=range(4))
+sos_run(['A', 'B'], idx=i)
+''')
+        import time
+        st = time.time()
+        wf = script.workflow()
+        # this should be ok.
+        Base_Executor(wf, config={'worker_procs': ['8']}).run()
+        self.assertTrue(time.time() - st < 20)
+
+    def testPassOfArgs(self):
+        '''Test passing of arguments through sos_run #1164'''
+        script = SoS_Script('''
+parameter: b=2
+[A]
+parameter: c=1
+print(c)
+
+[default]
+sos_run('A', a=2)
+''')
+        wf = script.workflow()
+        self.assertRaises(Exception, Base_Executor(wf).run)
+        #
+        script = SoS_Script('''
+parameter: b=2
+[A]
+parameter: c=1
+print(c)
+
+[default]
+sos_run('A', b=2)
+''')
+        wf = script.workflow()
+        Base_Executor(wf).run()
+        #
+        script = SoS_Script('''
+parameter: b=2
+[A]
+parameter: c=1
+print(c)
+
+[default]
+sos_run('A', c=2)
+''')
+        wf = script.workflow()
+        Base_Executor(wf).run()
+
+    def testNestedDynamicDepends(self):
+        '''Test the execution of nested workflow with dynamic depends'''
+        script = SoS_Script('''
+[A_1]
+parameter: num = 20
+input: f"B{num}.txt"
+depends: _input + '.p'
+
+[B: provides='B{num}.txt']
+run: expand=True
+    touch {_output}
+
+[P: provides='{filename}.p']
+input: filename
+run: expand=True
+    touch {_output}
 
 
+[default]
+sos_run('A', num=0)
+''')
+        wf = script.workflow()
+        Base_Executor(wf).run()
 
+    def testNesteWithBothWorkflowAndTargets(self):
+        '''Test nested workflow with both workflow and targets'''
+        for file in ('A_out.txt', 'B_out.txt'):
+            if os.path.isfile(file):
+                os.remove(file)
+        execute_workflow(r'''
+[A_20]
+output: 'A_out.txt'
+_output.touch()
+
+[B]
+output: 'B_out.txt'
+_output.touch()
+
+[default]
+sos_run('A', targets='B_out.txt')
+''')
+        self.assertTrue(os.path.isfile('A_out.txt'))
+        self.assertTrue(os.path.isfile('B_out.txt'))
 
 if __name__ == '__main__':
     #suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestParser)
-    #unittest.TextTestRunner(, suite).run()
+    # unittest.TextTestRunner(, suite).run()
     unittest.main()
